from ..domain import BaseEntity, DomainEvent

import pytest


class SampleEvent(BaseEntity):
    class Created(BaseEntity.Created):
        def mutate_event(self, event, klass):
            return klass(entity_id=event.entity_id,
                         entity_version=event.entity_version,
                         domain_event_id=event.domain_event_id)

    class Updated(DomainEvent):
        def mutate_event(self, event, entity):
            entity.is_dirty = True
            return entity

    class Closed(DomainEvent):
        # no mutate_event
        pass

    def __init__(self, is_dirty=False, **kwargs):
        super().__init__(**kwargs)
        self.is_dirty = is_dirty


def test_base_entity_calls_mutator():
    create_event = SampleEvent.Created(entity_id=1)
    entity = SampleEvent.mutate(event=create_event)

    close_event = SampleEvent.Updated(entity_id=entity.id, entity_version=entity.version)
    entity = SampleEvent.mutate(event=close_event, entity=entity)
    assert entity.is_dirty is True

def test_base_entity_requires_mutator():
    create_event = SampleEvent.Created(entity_id=1)
    entity = SampleEvent.mutate(event=create_event)

    close_event = SampleEvent.Closed(entity_id=entity.id, entity_version=entity.version)

    with pytest.raises(NotImplementedError):
        SampleEvent.mutate(event=close_event, entity=entity)


def test_create_for_event():
    event = SampleEvent.Created(
        entity_id='ENTITY_ID',
        domain_event_id='DOMAIN_EVENT_ID',
        entity_version=0,
    )
    obj = SampleEvent.create_for_event(event)

    assert obj.id == 'ENTITY_ID'
    assert obj.version == 0



from eventsourcing.domain.model.entity import DomainEvent
from eventsourcing.domain.model.entity import EventSourcedEntity
from eventsourcing.domain.model.entity import entity_mutator
from eventsourcing.domain.model.entity import singledispatch
from eventsourcing.domain.model.decorators import subscribe_to
from eventsourcing.domain.model.events import publish
from eventsourcing.domain.model.events import subscribe
from eventsourcing.domain.model.events import unsubscribe
from eventsourcing.infrastructure.event_sourced_repo import EventSourcedRepository
from .domain import BaseEntity
from .app import EventSourcingWithDjango

default_app_config = 'djangoevents.apps.AppConfig'

__all__ = [
    'DomainEvent',
    'EventSourcedEntity',
    'EventSourcedRepository',
    'entity_mutator',
    'singledispatch',
    'publish',
    'subscribe',
    'unsubscribe',
    'subscribe_to',
    'BaseEntity',
    'EventSourcingWithDjango'
]



from django.apps import AppConfig as BaseAppConfig
from django.conf import settings
from django.utils.module_loading import import_module
import os.path


class AppConfig(BaseAppConfig):
    name = 'djangoevents'

    def ready(self):
        for app_module_name in get_app_module_names():
            import_handlers_module(app_module_name)


def get_app_module_names():
    return settings.INSTALLED_APPS


def import_handlers_module(app_module_name):
    handlers_module_name = '%s.handlers' % app_module_name
    try:
        import_module(handlers_module_name)
    except ImportError:
        # we need to re-raise exception in case there was import errors inside
        # handlers.py module
        handlers_file_name = get_handlers_file_name(app_module_name)
        if os.path.exists(handlers_file_name):
            raise


def get_handlers_file_name(app_module_name):
    module = import_module(app_module_name)
    module_dir = os.path.dirname(module.__file__)
    return os.path.join(module_dir, 'handlers.py')



from eventsourcing.domain.model.entity import DomainEvent
from eventsourcing.domain.model.entity import EventSourcedEntity


class BaseEntity(EventSourcedEntity):
    """
    `EventSourcedEntity` with saner mutator routing:

    >>> class Asset(BaseEntity):
    >>>    class Created(BaseEntity.Created):
    >>>        def mutate(event, klass):
    >>>            return klass(...)
    >>>
    >>>    class Updated(DomainEvent):
    >>>        def mutate(event, instance):
    >>>            instance.connect = True
    >>>            return instance
    """
    @classmethod
    def mutate(cls, entity=None, event=None):
        if entity:
            entity._validate_originator(event)

        if not hasattr(event, 'mutate_event'):
            msg = "{} does not provide a mutate_event() method.".format(event.__class__)
            raise NotImplementedError(msg)

        entity = event.mutate_event(event, entity or cls)
        entity._increment_version()
        return entity

    @classmethod
    def create_for_event(cls, event):
        entity = cls(
            entity_id=event.entity_id,
            domain_event_id=event.domain_event_id,
            entity_version=event.entity_version,
        )
        return entity



from django.http import Http404


def get_entity_or_404(repo, entity_id):
    """
    Uses `repo` to return most recent state of the entity. Raises a Http404 exception if the
    `entity_id` does not exist.

    `repo` has to be a EventSourcedRepository object.
    """
    if entity_id not in repo:
        raise Http404
    return repo[entity_id]



"""
Same as settings_test but instead of migration
uses provided sql. This is for test to make sure
that Event model works fine with db scheme
all teams agreed on.
"""
from .settings_test import *
import os


DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql',
        'HOST': os.environ.get("MYSQL_HOSTNAME", '127.0.0.1'),
        'PORT': os.environ.get("MYSQL_PORT", 3306),
        'NAME': os.environ.get("MYSQL_DATABASE", 'djangoevents'),
        'USER': os.environ.get("MYSQL_USER", 'root'),
    }
}



from ..app import EventSourcingWithDjango
from ..models import Event
from ..repository import from_model_instance
from ..unifiedtranscoder import UnifiedStoredEvent
from eventsourcing.domain.model.entity import EventSourcedEntity
from eventsourcing.domain.services.transcoding import EntityVersion
from eventsourcing.utils.time import timestamp_from_uuid
from datetime import datetime
from django.db import transaction
from djangoevents.exceptions import AlreadyExists
import json
import pytest


class SampleAggregate(EventSourcedEntity):
    class Created(EventSourcedEntity.Created):
        pass


class SampleApp(EventSourcingWithDjango):
    def on_init(self):
        self.repo = self.get_repo_for_entity(SampleAggregate)


# uuid1 contains time
uuids = ['7ab23d4c-a520-11e6-80f5-76304dec7eb7',
         '846713a8-a520-11e6-80f5-76304dec7eb7',
         '917e42a0-a520-11e6-80f5-76304dec7eb7',
         '98ef1faa-a520-11e6-80f5-76304dec7eb7',
         '9e89b9de-a520-11e6-80f5-76304dec7eb7',
         'a678b7a8-a520-11e6-80f5-76304dec7eb7']


def new_stored_event(event_id, event_type, event_data, aggregate_id, aggregate_version):
    return UnifiedStoredEvent(
        event_id=event_id,
        event_type=event_type,
        event_data=json.dumps(event_data),
        aggregate_id=aggregate_id,
        aggregate_type='SampleAggregate',
        aggregate_version=aggregate_version,
        create_date=datetime.fromtimestamp(timestamp_from_uuid(event_id)),
        stored_entity_id='SampleAggregate::' + aggregate_id,
        metadata='',
        module_name='',
        class_name='',
    )


def save_events(repo, stored_events):
    for e in stored_events:
        repo.append(e)


@pytest.fixture
def repo():
    app = SampleApp()
    return app.stored_event_repo


@pytest.fixture
def stored_events():
    return [new_stored_event(uuids[i], 'Created', {'title': 'test'}, uuids[0], i) for i, uuid in enumerate(uuids)]


@pytest.mark.django_db
def test_repo_append(repo):
    test_stored_event = new_stored_event(uuids[0], 'Created', {'title': 'test'}, uuids[1], 1)

    # test basic save and get
    repo.append(test_stored_event)
    events = Event.objects.filter(stored_entity_id=test_stored_event.stored_entity_id)
    assert len(events) == 1
    assert test_stored_event == from_model_instance(events[0])

    # test unique constraint
    with transaction.atomic():
        test_stored_event2 = new_stored_event(uuids[2], 'Updated', {'title': 'updated test'}, uuids[1], 1)
        with pytest.raises(AlreadyExists):
            repo.append(test_stored_event2)

    # multiple saves
    test_stored_event3 = new_stored_event(uuids[2], 'Updated', {'title': 'updated test'}, uuids[1], 2)
    repo.append(test_stored_event3)
    events = Event.objects.filter(stored_entity_id=test_stored_event.stored_entity_id)
    assert len(events) == 2
    assert test_stored_event == from_model_instance(events[0])
    assert test_stored_event3 == from_model_instance(events[1])


@pytest.mark.django_db
def test_repo_get_entity_version(repo):
    test_stored_event = new_stored_event(uuids[0], 'Created', {'title': 'test'}, uuids[1], 1)
    repo.append(test_stored_event)
    entity_version = repo.get_entity_version(test_stored_event.stored_entity_id, 1)
    assert isinstance(entity_version, EntityVersion)
    assert entity_version.entity_version_id == 'SampleAggregate::' + uuids[1] + '::version::1'
    assert entity_version.event_id == uuids[0]


@pytest.mark.django_db
def test_repo_get_entity_events(repo, stored_events):
    save_events(repo, stored_events)

    # test get all
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id)
    assert retrieved_events == stored_events


@pytest.mark.django_db
def test_repo_get_entity_events_slicing(repo, stored_events):
    save_events(repo, stored_events)

    # test after
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id, after=stored_events[2].event_id)
    assert retrieved_events == stored_events[3:]

    # test until
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id, until=stored_events[2].event_id)
    assert retrieved_events == stored_events[:3]

    # test after + until
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id, after=stored_events[1].event_id,
                                              until=stored_events[4].event_id)
    assert retrieved_events == stored_events[2:5]

    # test limit
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id, limit=3)
    assert retrieved_events == stored_events[:3]


@pytest.mark.django_db
def test_repo_get_entity_events_ordering(repo, stored_events):
    save_events(repo, stored_events)

    # test query_ascending
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id, limit=3, query_ascending=False)
    assert retrieved_events == stored_events[-3:]

    # test results_ascending
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id, results_ascending=False)
    assert retrieved_events == stored_events[::-1]

    # test limit + query_ascending + results_ascending
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id, limit=3, query_ascending=False,
                                              results_ascending=False)
    assert retrieved_events == stored_events[-3:][::-1]

    # test after + until + query_ascending
    retrieved_events = repo.get_entity_events(stored_events[0].stored_entity_id, after=stored_events[1].event_id,
                                              until=stored_events[4].event_id, query_ascending=False)
    assert retrieved_events == stored_events[1:4]


@pytest.mark.django_db
def test_append_wrong_type(repo):
    # only allow UnifiedStoredEvent to be saved
    with transaction.atomic():
        with pytest.raises(AssertionError):
            repo.append('wrong object type')



from setuptools import setup

setup(
    name='djangoevents',
    version='0.9.4',
    url='https://github.com/ApplauseOSS/djangoevents',
    license='MIT',
    description='Building blocks for building Event Sourcing Django applications.',
    author='Applause',
    author_email='eng@applause.com',
    zip_safe=False,
    packages=[
        'djangoevents',
        'djangoevents.migrations',
        'djangoevents.tests.settings',
    ],
    include_package_data=True,
    install_requires=[
        'eventsourcing>=1.2,<1.3',
        'django',
    ],
)
